@@ -11,4 +11,6 @@ from PDSUtilities.plotly.utilities import hex_to_rgb
 from PDSUtilities.plotly.utilities import rgb_to_hex
 from PDSUtilities.plotly.utilities import get_colors
 from PDSUtilities.plotly.utilities import update_width_and_height
-from PDSUtilities.plotly.utilities import update_title
\ No newline at end of file
+from PDSUtilities.plotly.utilities import update_title
+#
+from PDSUtilities.plotly.create_image_subplots import create_image_subplots
